// *****************************************************************************
// * IBM MQSeries Workflow XML Sample Scenario                                 *
// *****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999-2000
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
// *****************************************************************************
// *****************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
// *****************************************************************************
package com.ibm.workflow.upes;

import java.lang.String;
import java.util.Vector;


/**
 * The <b>ContainerElement</b> class is a simple representation
 * of MQSeries Workflow container elements. It has the ability to navigate through
 * a TXElement based 'XML parse tree' and to get the
 * container element information out of an XML container representation.
 * On the other side it provides an interface to access its value
 * or its subelements. It also provides an interface to generate an XML representation
 * of itself.
 * A container element either contains a value represented as string or a Vector of subelements.
 * I.e. at least one of both fields (_Value and _Children is always null).
 * @version 1.0
 */
public class ContainerElement
{
  /************************************************************
   * <b>_Name</b> is the Name of the container element.
   * Every element has a name. This name is use as and retrieved from the XML tag name.
   * @see String
   * @see #GetName
   *
  ************************************************************/
  private String _Name;
  /************************************************************
   * <b>_Value</b> is the string representation of the container elements value.
   * A container element either contains a value represented as string or a Vector of subelements.
   * I.e. at least one of both fields (_Value and _Children is always null).
   * @see String
   * @see #GetValue 
   * @see #SetValue
   *
  ************************************************************/
  private String _Value;
  /************************************************************
   * <b>_Children</b> is the Vector containing the sub container elements of the given container element.
   * A container element either contains a value represented as string or a Vector of subelements.
   * I.e. at least one of both fields (_Value and _Children is always null).
   * @see Vector
   * @see #GetChildren
   * @see #AddChild
   *
  ************************************************************/
  private Vector _Children;

/************************************************************
 * Constructor which set the name of the container element.
 * @parameters <b>name</b> the name of the container element.
 * 
 ***********************************************************/
  public ContainerElement(String name)
  {
    _Name = name;
  }
/************************************************************
 * <b>AddChild</b> is the method to add a sub container element to the given element.
 * It simply appends the handed over container element.
 * @parameters <b>c</b> is the container element which is to be appended to the vector.
 * 
 ***********************************************************/
  public void AddChild(ContainerElement c)
  {
    if (_Children == null)
    {
      //value must be set to null
      _Value = null;
      _Children = new Vector();
    }
    _Children.addElement(c);
  }
/************************************************************
 * <b>RemoveChild</b> is the method to delete a container element from the given elements' children.
 * If the container element <b>c</b> was not a child nothing happens.
 * @parameters <b>c</b> is the container element which is to be appended to the vector.
 * 
 ***********************************************************/
  public void RemoveChild(ContainerElement c)
  {
    if (_Children != null)
    {
      _Children.removeElement(c);
    }
  }
/************************************************************
 * <b>AsXMLString</b> returns an XML representation of the given container element,
 * taking the indention in consideration. Optionally it is possible to suppress the
 * output of tags (this is used for root elements of containers because they have only
 * subelements and the root element itself is not expressed with tags in the XML output.
 *
 * @parameters <b>indention</b> is the prefix used for each line of the output.
 * @parameters <b>withTags</b> is a flag indication if tag names are to be put out too.
 * @return  XML representation of the given container element
 * 
 ***********************************************************/
  public String AsXMLString(String indention, boolean withTags)
  {
    String result = "";
    if ((_Value == null) && (_Children == null))
    {//the container element is empty 
      if (withTags)
      {
        result = indention+"<"+_Name+"/>";
      }
    }
    else
    {//the container element is not empty
      if (withTags)
      {
        result = indention+"<"+_Name+">";
      }
      if (_Value != null)
      {
        result = result + _Value;
      }
      else
      {
        //all subelements must be put out in their order they appear in the vector
        for (int i = 0; i < _Children.size(); i++)
        {
          result = result + "\n" + ((ContainerElement)_Children.elementAt(i)).AsXMLString(indention+"  ",true);
        }
        result = result+"\n" + indention;
      }
      if (withTags)
      {
        result = result +"</"+_Name+">";
      }
    }
    return result;
  }
  /************************************************************
  * <b>GetChildNamed</b> returns the first sub container element which has the given name.
  * @parameters <b>name</b> of the wanted subelement
  * @return  the first sub container element which has the given name.
  * <p><b>null</b> if there is no sub container element with the given name.
  * 
  ***********************************************************/
  public ContainerElement GetChildNamed(String name)
  {
    if (_Children != null)
    {
      for (int i = 0; i < _Children.size (); i++)
      {
        if (((ContainerElement)_Children.elementAt(i))._Name.equals(name))
        {
          return(ContainerElement)_Children.elementAt(i);
        }
      }
    }
    return null;
  } 
/************************************************************
 * <b>SetValue</b> setter of _Value.
 * _Children are set to null because _Value and _Children exclude each other.
 * @parameters <b>value</b> is the string value to be stored in _Value
 * 
 ***********************************************************/
  public void SetValue(String value)
  {
    _Value = value;
    _Children = null;
  }
/************************************************************
 * <b>GetValue</b> getter of _Value.
 * returns the content of _Value.
 * @return the content of _Value.
 * 
 ***********************************************************/
  public String GetValue()
  {
    return _Value;
  }
/************************************************************
 * <b>GetChildren</b> getter of _Children, the vector of sub container elements.
 * @return the content of _Children, the vector of sub container elements.
 * 
 ***********************************************************/
  public Vector GetChildren()
  {
    return _Children;
  }
/************************************************************
 * <b>GetValue</b> getter of _Value.
 * returns the content of _Value.
 * @return the content of _Value.
 * 
 ***********************************************************/
  public String GetName()
  {
    return _Name;
  }

  public ContainerElement GetUserData()
  {
    if ( _Children != null)
    {
      //- skip pre-defined data members
      for ( int i=0; i < _Children.size(); i++)
      {
        ContainerElement tempEl = (ContainerElement) _Children.get( i);
        if ( !tempEl.GetName().startsWith( "_"))
        {
          //- got it
          return tempEl;
        }
      }
    }
    return null;
  }  


}
