// *****************************************************************************
// * IBM MQSeries Workflow XML Sample Scenario                                 *
// *****************************************************************************
//
//  Licensed Materials - Property of IBM
//  5697-FM3 5655-A96
//  (C) Copyright IBM Corp. 1999-2000
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
// *****************************************************************************
// *****************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
// *****************************************************************************

package com.ibm.workflow.upes;

import com.ibm.mq.*;            // Include the MQSeries classes for Java package
import java.util.Hashtable;


/**
 * The <code>MQHandler</code> encapsulates the MQSeries API to what is needed
 * by the WFMessageAdapter.
 * For performance reasons it keeps the handle of opened queues and queue
 * managers.
 *
 * @author Frank Skrzypczak
 * @version 1.0.0
 */
public class MQHandler
{
   /**
    * The handle of the local queue manager
    */
   private MQQueueManager _localQueueManager;
   /**
    * List of queue handles the WFMessageAdapter put messages to
    */
   private Hashtable _queues4Put = new Hashtable();
   /**
    * List of queue handles the WFMessageAdapter get messages from
    */
   private Hashtable _queues4Get = new Hashtable();

   /**
    * Opens a new queue manager object and keeps its handle.
    * Note: The handle is kept for performance reasons               
    * 
    * @param queueManagerName The name of the local queue manager
    * @exception com.ibm.mq.MQException
    */
   public MQHandler( String queueManagerName) throws com.ibm.mq.MQException
   {
      _localQueueManager = new MQQueueManager(queueManagerName); 
   }
   /**
    * Transforms the payload of a MQSeries message into a JAVA string.
    * Note that MQSeries Workflow messages are encoded in UTF-8 while
    * a JAVA string is encoded in UNICODE
    * 
    * @param mq     The MQSeries message
    * @return The transformed payload of the message.
    * @exception java.io.IOException
    */
   public String transformMessage( MQMessage mq) throws java.io.IOException
   {
      byte b[] = new byte[ mq.getDataLength( ) ];
      mq.readFully( b, 0, mq.getDataLength( ) );
      String inputXMLMessage = new String( b ); 
      return inputXMLMessage;
   }
   /**
    * Get a message from an input queue. Wait unlimited until a message
    * arrives.
    * Input queue and queue manager handles are cached
    * 
    * @param queueManagerName
    *                  The input queue manager name
    * @param queueName The input queue name
    * @return MQMessage
    * @exception com.ibm.mq.MQException
    */
   public MQMessage getMessage( String queueManagerName, 
                                String queueName) throws com.ibm.mq.MQException
   {
      MQQueue getQueue = getQueue4Get( queueManagerName,
                                       queueName);

      // First define a MQSeries message buffer to receive the message into..
      MQMessage retrievedMessage = new MQMessage();

      // Set the get message options...
      MQGetMessageOptions gmo = new MQGetMessageOptions(); 
      // with syncpoit protocol 
      gmo.options = MQC.MQGMO_WAIT | MQC.MQGMO_SYNCPOINT;
      gmo.waitInterval = MQC.MQWI_UNLIMITED;
      // get a message transactional from the queue 
      Tracer.trace("Get message from " + queueManagerName + ":" + queueName);
      getQueue.get(retrievedMessage, gmo);
      return retrievedMessage;
   }
   /**
    * Get a message from an input queue. Don't wait if there is no message.
    * Input queue and queue manager handles are cached
    * 
    * @param queueManagerName
    *                  The input queue manager name
    * @param queueName The input queue name
    * @return MQMessage
    * @exception com.ibm.mq.MQException
    */
   public MQMessage getMessageNoWait( String queueManagerName, 
                                String queueName) throws com.ibm.mq.MQException
   {
      MQQueue getQueue = getQueue4Get( queueManagerName,
                                       queueName);

      // First define a MQSeries message buffer to receive the message into..
      MQMessage retrievedMessage = new MQMessage();

      // Set the get message options...
      MQGetMessageOptions gmo = new MQGetMessageOptions(); 
      // with syncpoit protocol 
      gmo.options = MQC.MQGMO_NO_WAIT | MQC.MQGMO_SYNCPOINT;
      gmo.waitInterval = MQC.MQWI_UNLIMITED;
      // get a message transactional from the queue 
      Tracer.trace("Get message from " + queueManagerName + ":" + queueName);
      getQueue.get(retrievedMessage, gmo);
      return retrievedMessage;
   }
   /**
    * Put a message into an output queue. 
    * Output queue and queue manager handles are cached
    * 
    * @param message2Put
    *                  The MQSeries message to put
    * @param queueManagerName
    *                  The output/target queue manager name
    * @param queueName The output/target queue name
    *                  
    * @return true
    * @exception com.ibm.mq.MQException
    */
   public boolean putMessage( MQMessage message2Put,
                              String    queueManagerName, 
                              String    queueName) throws com.ibm.mq.MQException
   {
      MQQueue putQueue = getQueue4Put( queueManagerName,
                                   queueName);

      // specify the message options...
      MQPutMessageOptions pmo = new MQPutMessageOptions();
      // with syncpoit protocol 
      pmo.options = MQC.MQPMO_SYNCPOINT;

      // put the message transactional into the queue
      Tracer.trace("Put message to " + queueManagerName + ":" + queueName);
      putQueue.put(message2Put,
                   pmo);
      Tracer.trace("put completed");
      return true;
   }

   /**
    * Get a queue handle for a MQSeries GET operation
    * 
    * @param queueManagerName
    * @param queueName
    * @return A MQ queue handle
    * @exception com.ibm.mq.MQException
    */
   public MQQueue getQueue4Get( String queueManagerName, 
                                String queueName) throws com.ibm.mq.MQException
   {
      MQQueue queue;

      if ( !_queues4Get.isEmpty() && _queues4Get.containsKey( queueName))
      {
         queue = (MQQueue) _queues4Get.get( queueName);
      }
      else
      {
         int openOptions = MQC.MQOO_INPUT_SHARED;

         // Now open the queue where we expect the response....
         Tracer.trace("Access Queue: " + queueName + " at Queue Manager:" + queueManagerName);

         queue = _localQueueManager.accessQueue(queueName,
                                                openOptions,
                                                queueManagerName,
                                                null,
                                                null
                                                );

         _queues4Get.put( queueName, queue);
      }
      return queue;

   }
   /**
    * Get a queue handle for a MQSeries PUT operation
    * 
    * @param queueManagerName
    * @param queueName
    * @return A MQ queue handle
    * @exception com.ibm.mq.MQException
    */
   public MQQueue getQueue4Put( String queueManagerName, 
                                String queueName) throws com.ibm.mq.MQException
   {
      MQQueue queue;
      if ( !_queues4Put.isEmpty() && _queues4Put.containsKey( queueName))
      {
         queue = (MQQueue) _queues4Put.get( queueName);
      }
      else
      {
         int openOptions = MQC.MQOO_OUTPUT ;

         // Now open the queue where we put a message....
         Tracer.trace("Access Queue: " + queueName + " at Queue Manager:" + queueManagerName);
         queue = _localQueueManager.accessQueue(queueName,
                                                openOptions,
                                                queueManagerName,
                                                null,
                                                null
                                                );

         _queues4Put.put( queueName, queue);
      }
      return queue;

   }

   /**
    * Commit the transaction
    * 
    * @return true if operation was successful
    * @exception com.ibm.mq.MQException
    */
   public boolean commit() throws com.ibm.mq.MQException
   {
     Tracer.trace("commit on queue manager");
     if (_localQueueManager != null)
     {
       _localQueueManager.commit();
       Tracer.trace("commit completed successfully");
       return true;
     }
     else
     {
       return false;
     }
   }

   /**
    * Rollback the transaction
    * 
    * @return true if operation was successful
    * @exception com.ibm.mq.MQException
    */
   public boolean rollback() throws com.ibm.mq.MQException
   {
     Tracer.trace("Rollback on queue manager");
     if (_localQueueManager != null)
     {
       _localQueueManager.backout();
       Tracer.trace("commit completed rollback");
       return true;
     }
     else
     {
       return false;
     }
   }
}
