/*BEGIN CMVC */
//*************************************************************************
//
//  Workfile: %W%
//  Last update: %E% %U%
//  SCCS path, id: %F% %I%
//
//************************************************************************/
/*END CMVC */
/*BEGIN COPYRIGHT */
//*************************************************************************
//
//  Licensed Materials - Property of IBM
//  WA07 Support Pac - Web Services Process Management Toolkit
//  (c) Copyright IBM Corp. 2001
//  US Government Users Restricted Rights - Use, duplication or
//  disclosure restricted by GSA ADP Schedule Contract
//  with IBM Corp.
//
//************************************************************************/
/*END COPYRIGHT */
/*BEGIN DISCLAIMER */
//*************************************************************************
//  DISCLAIMER
//  ----------
//  This material contains programming source code for your consideration.
//  These examples have not been thoroughly tested under all conditions.
//  IBM, therefore, cannot guarantee or imply reliability, serviceability,
//  or function of these programs.
//  ALL PROGRAMS CONTAINED HEREIN ARE PROVIDED TO YOU "AS IS", WITHOUT ANY
//  WARRANTIES (EXPRESS OR IMPLIED) OR SUPPORT WHATSOEVER, INCLUDING BUT
//  NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS
//  FOR A PARTICULAR PURPOSE.
//************************************************************************/
/*END DISCLAIMER */

package com.ibm.workflow.util.fdl.make;

import com.ibm.workflow.util.fdl.*;
import java.io.PrintWriter;
import java.util.*;

/**
 *
 * <p>An FDL parse tree visitor which collects the information about all FDL STRUCTURE's in the tree.
 * This data is then used by the code-generation methods (<code>generateXXX</code>) to map FDL STRUCTURE's
 * to their respective target representation.</p>
 * <p>This visitor operates on a parse tree which was generated from
 * <a href="../../../../../../../fdlgrammar.html">this</a>
 * FDL grammar using the <b>JavaCC parser generator</b>. It operates on the sub-tree starting at the
 * <a href="../../../../../../../fdlgrammar.html#prod12">Structure</a>
 * production.</p>
 *
 * @author  Robert Junghuber (<a href='mailto:junghub@de.ibm.com'>junghub@de.ibm.com</a>)
 * @version 1.2
 *
 */

public class StructureVisitor extends NopVisitor
{
  // Helper classes
  // --------------

  protected abstract class Setting { String name; } // All FDL stettings have a name.

  protected class Member extends Setting // Represents an FDL structure member
  {
    // Generate COBOL copybook definition for FDL structure member.

    public void toCPYDefinition() throws ParseException
    {
      pw.print(indent() + name + (card > 0 ? (" OCCURS " + card + " TIMES") : ""));

      if (isType())
      {
        // Recursive procedure to flatten member. Inline-expand nested structure. Isn't COBOL wonderful?

        pw.println(".");

        for (int i = 0; i < structures.size(); i++)
        {
          Structure s = (Structure)structures.get(i);

          if (s.name.equals(type)) { incLevel(s.name); s.toCPYDeclaration(); cpyLevel--; break; }
        }
      }
      else
      {
        // Down to simple type. Generate what's configured.

        pw.print(" ");

        switch(kind)
        {
          case Parser.BINARY : pw.print(cpyType[0]); break;
          case Parser.LONG   : pw.print(cpyType[1]); break;
          case Parser.FLOAT  : pw.print(cpyType[2]); break;
          case Parser.STRING : pw.print(cpyType[3]); break;
        }

        pw.println(".");
      }
    }

    // Generate DTD declaration for FDL stucture member.

    public void toDTDDeclaration()
    {
      pw.print(name);

      // Member elements are optional if they are nested structures, occur zero or more times when they are structure
      // arrays, and are optional if they are simple types.

      if      (isType())  pw.print(card == 0 ? '?' : '*');
      else if (card == 0) pw.print('?');
    }

    // Generate DTD definition for FDL structure member. This generates an ELEMENT which either refers to the corresponding
    // ENTITY (in case this member has a STRUCTURE type) or to the pre-defined ENTITIES representing simple types.

    public void toDTDDefinition()
    {
      if (!skip) pw.println("<!ELEMENT " + name +
                            (isType() ? " (%" + type.toUpperCase() + ";)"
                                      : " %" + type + (card == 0 ? ";" : "_ARRAY;")) +
                            ">" + eol);
    }

    // Generate XML Schema declaration for FDL structure member. Note that unlike DTD's there is no need for a
    // corresponding definition.

    public void toXSDDeclaration()
    {
      // Member elements occur (0..1) if they are nested structures, (0..card) times if they are structure arrays,
      // and (0..1) if they are simple types.

      pw.print("  <xs:element name='" + name + "' type='");

      if      (isType()) pw.print("%SNS%:" + type + "' minOccurs='0' maxOccurs='" + (card == 0 ? 1 : card));
      else if (card > 0) pw.print((doNull ? ("%WFNS%:" + type + "_ARRAY") : ("xs:" + type.toLowerCase() + "' minOccurs='0' maxOccurs='unbounded")));
      else               pw.print("%WFNS%:" + type + "' minOccurs='0' maxOccurs='1");

      pw.println("'/>");
    }

    // Answer true if this member has none of the FDL basic types (i.e. 'kind' doesn't match a basic type FDL keyword).

    private boolean isType() { return kind == Parser.STRING_LITERAL || kind == Parser.IDENTIFIER; }

    String  type; // FDL type name of the member
    int     kind; // Kind of the member (as established by the FDL parser, helps avoiding string comparisons on 'type')
    int     card; // Cardinality of the member, zero for scalar
    boolean skip = false; // Another compatible member exists in another structure, prevent duplicate DTD definition
  }

  protected class Structure extends Setting // Represents an FDL structure
  {
    // Generate COBOL copybook declaration for FDL structure. Simply enumerates all members.

    public void toCPYDeclaration() throws ParseException
    {
      for (Iterator it = members.iterator(); it.hasNext() ;) ((Member)it.next()).toCPYDefinition();
    }

    // Generate COBOL copybook decfinition for FDL structure. This is the top-level entry point.

    public void toCPYDefinition() throws ParseException
    {
      if (skip) return;

      pw.println(indent() + name + ".");

      incLevel(name); toCPYDeclaration(); cpyLevel--;

      pw.print(eol);
    }

    // Generate DTD declaration for FDL structure.

    public void toDTDDeclaration() { if (!skip) pw.print(name); }

    // Generate DTD definition for FDL structure.

    public void toDTDDefinition()
    {
      if (skip) return;

      // First, generate ELEMENT definitions for all members. They all share one namespace, that's why we validate against
      // a flat namespace in generateDTD().

      for (Iterator it = members.iterator(); it.hasNext() ;) ((Member)it.next()).toDTDDefinition();

      // Generate an ENTITY unless the structure has no members

      if (members.size() > 0)
      {
        pw.print("<!ENTITY % " + name.toUpperCase());

        if (members.size() > 1) pw.print(eol);

        pw.print(" '");

        int cnt = members.size();

        // Let each member generate its DTD declaration.

        for (Iterator it = members.iterator(); it.hasNext() ;)
        {
          ((Member)it.next()).toDTDDeclaration();

          if (--cnt > 0) pw.print(',' + eol + "  ");
        }

        pw.print("'");

        if (members.size() > 1) pw.print(eol);

        pw.println(">" + eol);
      }

      // Generate an ELEMENT, taking care of empty structures

      pw.println("<!ELEMENT " + name + (members.size() == 0 ? " EMPTY>" : (" (%" + name.toUpperCase() + ";)>")) + eol);
    }

    // Generate XML Schema declaration for FDL structure. This maps to an <element> with the corresponding <complexType>.

    public void toXSDDeclaration() { if (!skip) pw.println("   <xs:element name='" + name + "' type='%SNS%:" + name + "'/>"); }

    // Generate XML Schema definition for FDL structure. Note that unlike DTD there is no need to generate a corresponding
    // top-level <element>.

    public void toXSDDefinition()
    {
      if (skip) return;

      // Generate a <complexType> for each structure.

      pw.print("<xs:complexType name='" + name + "'");

      if (members.size() > 0)
      {
        pw.println(">" + eol +" <xs:sequence>");

        // Let each member generate its XML Schema declaration.

        for (Iterator it = members.iterator(); it.hasNext() ;) ((Member)it.next()).toXSDDeclaration();

        pw.println(" </xs:sequence>" + eol + "</xs:complexType>" + eol);
      }
      else pw.println("/>" + eol); // Typically, only the DefaultDataStructure will be 'empty'.
    }

    // Append the types self depends on (including self as last element) in an orderly fashion to <code>result</code>.

    public void getClosure(List v)
    {
      for (int i = 0; i < members.size(); i++)
      {
        Member m = (Member)members.get(i);

        if (m.isType())
        {
          for (int j = 0; j < structures.size(); j++)
          {
            Structure s = (Structure)structures.get(j);

            if (s.name.equals(m.type)) { s.getClosure(v); break; }
          }
        }
      }

      if (members.size() > 0 && !v.contains(this)) v.add(this);
    }

    List    members = new ArrayList(10); // Ordered list of members
    boolean skip    = false;             // If true, Structure should be skipped during generation
  }

  /**
   *
   * Generates COBOL copybook declarations. FDL data type mapping, indentation spacing and level numbering are defined
   * by the corresponding configuration settings.
   *
   */

  public void generateCPY() throws ParseException
  {
    // Read configuration parameters.

    Hashtable types = (Hashtable)settings.get("TYPES");

    cpyType[0] = (String)types.get("BINARY");
    cpyType[1] = (String)types.get("LONG"  );
    cpyType[2] = (String)types.get("FLOAT" );
    cpyType[3] = (String)types.get("STRING");

    Hashtable nesting = (Hashtable)settings.get("NESTING");

    ind = Integer.parseInt((String)nesting.get("Indentation"));
    inc = Integer.parseInt((String)nesting.get("Increment"));

    // Trigger copybook generation for all structures

    for (Iterator it = structures.iterator(); it.hasNext() ;) ((Structure)it.next()).toCPYDefinition();
  }

  /**
   *
   * Generates Data Type Definition. Performs name uniqueness validation as a side-effect.
   *
   */

  public void generateDTD() throws ParseException
  {
    // Verify the name uniqueness of the parsed structures and members against the message elements.

    verifyDTD();

    // Trigger DTD generation for all structures

    for (Iterator it = structures.iterator(); it.hasNext() ;) ((Structure)it.next()).toDTDDefinition();

    // Generate ENTITY representing all container alternatives; the union is prefixed by the _CONTAINER_INFO ENTITY.

    int cnt = 0;

    pw.print("<!ENTITY % _CONTAINER" + eol + " '%_CONTAINER_INFO;" + (structures.size() > skipCnt ? "," : ""));

    if (structures.size() > skipCnt + 1) pw.print(eol + " ( ");

    for (Iterator it = structures.iterator(); it.hasNext() ;)
    {
      pw.print(eol + (cnt++ == 0 ? "   " : " | "));

      ((Structure)it.next()).toDTDDeclaration();
    }

    if (structures.size() > skipCnt + 1) pw.print(eol + " )'");

    pw.println(eol + ">");
  }

  /**
   *
   * Generates XML Schema.
   *
   */

  public void generateXSD() throws ParseException
  {
    // Trigger XML Schema generation for all structures

    for (Iterator it = structures.iterator(); it.hasNext() ;) ((Structure)it.next()).toXSDDefinition();

    // Generate a complexType representing all container alternatives; the union is prefixed by the _CONTAINER_INFO group.

    pw.println("<xs:complexType name='_CONTAINER'>" + eol + " <xs:sequence>" + eol +"  <xs:group ref='%WFNS%:_CONTAINER_INFO'/>");

    if (structures.size() > skipCnt + 1) pw.println("  <xs:choice>");

    for (Iterator it = structures.iterator(); it.hasNext() ;) ((Structure)it.next()).toXSDDeclaration();

    if (structures.size() > skipCnt + 1) pw.println("  </xs:choice>");

    pw.println(" </xs:sequence>" + eol + "</xs:complexType>");
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod12">Structure</a>.
   *
   */

  public Object visit(FdlStructure node, Object data) throws ParseException
  {
    scope = new Structure(); // Open a new Structure to receive the data from child node visits.

    node.childrenAccept(this, Boolean.TRUE); // Recurse into child nodes, data != null un-NOP's FdlName visits.

    structures.add(scope); // Add the Structure to structures

    scope = null;

    return data;
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod94">MemberDeclaration</a>.
   *
   */

  public Object visit(FdlMemberDeclaration node, Object data) throws ParseException
  {
    SimpleNode n = (SimpleNode)(node.jjtGetChild(1));

    // FDL grammar allows comma-seperated member declaration list, but this is a pain to visit. Since FDL export doesnt
    // generate that anyway, it's simply detected and blocked.

    if (n.id() == Parser.JJTNAME)
      throw new ParseException("Comma-separated member declaration in STRUCTURE '"
                               + scope.name
                               + "' ( ..., "
                               + n.getFirstToken().image
                               + ") is not supported."
                               + eol
                               + "Please split into multiple declarations.");

    Structure parent = (Structure)scope; // The Structure this MemberDeclaration belongs to.

    scope = new Member(); // Open a new Member to receive the data from child node visits.

    node.childrenAccept(this, data); // Recurse into child nodes.

    parent.members.add(scope); // Add Member to Structure.

    scope = parent; // Fix scope for following MemberDeclaration visits.

    return data;
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod96">MemberCardinality</a>.
   *
   */

  public Object visit(FdlMemberCardinality node, Object data) throws ParseException
  {
    // MemberCardinality is the value of the array modifier of a MemberDeclaration.

    ((Member)scope).card = Integer.parseInt(node.getFirstToken().next.image); return data;
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod95">MemberType</a>.
   *
   */

  public Object visit(FdlMemberType node, Object data) throws ParseException
  {
    // A MemberType is either a simple type (in this case the visited node has no children) or the name of another
    // structure (in this case it is the first token of the first [and only] child node).

    Token  t = (node.jjtGetNumChildren() == 0 ? node : (SimpleNode)node.jjtGetChild(0)).getFirstToken();
    Member m = (Member)scope;

    m.type = removeBlanks(t.image);
    m.kind = t.kind; // If this is Parser.IDENTIFIER/STRING_LITERAL, then MemberType refers to another structure.

    return data;
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod28">Name</a>.
   *
   */

  public Object visit(FdlName node, Object data) throws ParseException
  {
    if (data != null) // data serves as NOP-flag
    {
      // FDL allows for structure names to contain spaces, DTD/XML does not. So all names are stripped from spaces.

      String n = removeBlanks(node);

      // Just being picky; if a name is repeated in the END statement, we make sure that it matches.

      if (scope.name != null && !scope.name.equals(n))
        throw new ParseException("Different END identifier '" + n + "' in declaration for STRUCTURE '" + scope.name + "'");

      scope.name = n; // Whatever the current scope is (either structure or member), now it has a name.
    }

    return data;
  }

  /**
   *
   * Visit an FDL <a href="../../../../../../../fdlgrammar.html#prod29">ProcessingOption</a>. Simply ignored.
   *
   */

  public Object visit(FdlProcessingOption node, Object data) throws ParseException { return data; }

  // Helper methods
  // --------------

  /**
   *
   * Strip all blanks from an FdlName
   *
   */

  protected final String removeBlanks(FdlName node) { return removeBlanks(node.getFirstToken().image); }

  private final String removeBlanks(String s)
  {
    StringBuffer tmp = new StringBuffer(s.length());

    for (int i = 0; i < s.length(); i++)
    {
      char c = s.charAt(i);

      if (c != ' ') tmp.append(c);
    }

    return tmp.toString();
  }

  // Increment nesting level for a copybook entry

  private void incLevel(String name) throws ParseException
  {
    if (++cpyLevel >= 50) throw new ParseException("Nesting level in STRUCTURE '" + name + "' is too deep.");
  }

  // Answer the indentation string according to the current nesting level (and the configured number of spaces per level
  // plus the nesting 'number' for the copybook entry (using the configured numbering increment).

  private String indent()
  {
    int lvl = inc * cpyLevel; if (lvl == 0) lvl++;

    return spc.substring(0, ind * cpyLevel) + (lvl < 10 ? "0" : "") + (lvl) + " ";
  }

  // Verify that the structure names and the structure member names do not collide with the element names used in the
  // invariant part of the MQWF XML messages. Note that as a side-effect the skip-flag in Structure instances is set
  // if this is indicated by the configuration. Also, duplicate but compatible members in different structures are flagged
  // to prevent duplicate element definition.

  private void verifyDTD() throws ParseException
  {
    // 'names' is a dictionary which contains:
    //
    // 1) all reserved names  (key = reserved name, value = false)
    // 2) all structure names (key = structure name, value = true)
    // 3) all member names    (key = member name, value = owning structure)

    names.clear();

    for (int i = 0; i < elements.length; i++) names.put(elements[i], Boolean.FALSE); // start out with reserved names only

    StringBuffer report =
      new StringBuffer("The FDL source does not conform to the constaints imposed for DTD generation." + eol +
                       "Unless you remove the following collision(s), the generated file would contain" + eol +
                       "invalid duplicate ELEMENT definitions:" + eol + eol);
    int errors = 0;

    for (Iterator ci = structures.iterator(); ci.hasNext() ;)
    {
      // Check the structure name ...

      Structure s = (Structure)ci.next();

      if (skipList != null) s.skip = "yes".equalsIgnoreCase((String)skipList.get(s.name));

      if (s.skip) { System.out.println("Skipping STRUCTURE '" + s.name + "'"); skipCnt++; continue; }

      Object o = names.get(s.name);

      if (o instanceof Structure) // have seen a member with the same name
      {
        report.append(++errors + ") STRUCTURE name '" + s.name + "' collides with member name in STRUCTURE '" + ((Structure)o).name + "'" + eol);
      }
      else if (o instanceof Boolean) // either a duplicate structure name or a collision with a reserved element name
      {
        if (((Boolean)o).booleanValue()) report.append(++errors + ") Duplicate STRUCTURE name '" + s.name + "'" + eol);
        else                             report.append(++errors + ") STRUCTURE name '" + s.name + "' collides with reserved ELEMENT name" + eol);
      }
      else names.put(s.name, Boolean.TRUE); // no collision yet, remember structure name

      // ... and all the member names

      for (Iterator mi = s.members.iterator(); mi.hasNext() ;)
      {
        Member m = (Member)mi.next();

        o = names.get(m.name);

        String msg = null;

        if (o == null)
        {
          names.put(m.name, s); // no collision yet, remeber member name and owning structure

          continue;
        }
        else if (o instanceof Structure) // potential collision with another member, check type compatibility
        {
          List v = ((Structure)o).members;

          for (int i = 0; i < v.size(); i++)
          {
            Member other = (Member)v.get(i);

            if (!other.name.equals(m.name)) continue;

            if (other.type.equals(m.type) && other.card == m.card) m.skip = true; // members are compatible, skip one
            else
            {
              msg = "member name in STRUCTURE '" + ((Structure)o).name + "'";

              break;
            }
          }
        }
        else if (o instanceof Boolean) // Member collides with another structure's name or with a reserved element name
        {
          boolean b = ((Boolean)o).booleanValue();

          if (b && m.name.equals(m.type)) // member type matches structure name --> same ELEMENT definition --> no collision
          {
            m.skip = true; // but we better generate the element only once

            continue;
          }

          msg = b ? "STRUCTURE name" : "reserved ELEMENT name";
        }

        if (msg != null) report.append(++errors + ") Member name '" + m.name + "' in STRUCTURE '" + s.name +
                                       "' collides with " + msg + eol);
      }
    }

    if (errors != 0) throw new ParseException(report.toString());
  }

  /**
   *
   * Construct a Visitor, passing configuration information for later access and the writer to the output stream
   *
   */

  public StructureVisitor(Make cxt)
  {
    this.cxt      = cxt;
    this.pw       = cxt.getPrintWriter();
    this.settings = cxt.getConfig();

    // Read skip list from configuration, if any

    if (settings != null)
    {
      skipList = (Hashtable)settings.get("SKIP");

      Hashtable options = (Hashtable)settings.get("OPTIONS");

      if (options != null)
      {
//      doNull = "yes".equals((String)options.get("EnableNull"));    // (JRNYI)
      }
    }
  }

  // Members
  // -------

  protected Make        cxt        = null;
  protected PrintWriter pw         = null;
  protected List        structures = new ArrayList(100);
  protected boolean     doNull     = true;
  private   Hashtable   settings   = null;
  private   Hashtable   skipList   = null;
  private   int         skipCnt    = 0;
  private   Setting     scope      = null;
  private   int         cpyLevel   = 0;
  private   int         ind        = 0;
  private   int         inc        = 0;
  private   String[]    cpyType    = new String[4];

  private  static final String eol = System.getProperty("line.separator", "\n");
  private  static final String spc = "                                                                                     ";
  private  static final String[] elements =
  {
    "ActImplCorrelID",
    "ActivityExpired",
    "ActivityImplInvoke",
    "ActivityImplInvokeResponse",
    "ActivityName",
    "ActivityState",
    "ActivityType",
    "AssociatedObject",
    "AuditEvent",
    "AuditTrailRecord",
    "AutomaticClose",
    "BackwardMappingFormat",
    "BackwardMappingParameters",
    "BlockNames",
    "Category",
    "CodePage",
    "CommandParameters",
    "CoordinatorOfRole",
    "Deep",
    "DllOptions",
    "DllV2Compatible",
    "Duration",
    "Duration2",
    "EntryPointName",
    "Environment",
    "Exception",
    "ExeOptions",
    "ExecutableName",
    "ExecutableType",
    "ExecuteFenced",
    "ExternalOptions",
    "ExternalProcessContext",
    "ForwardMappingFormat",
    "ForwardMappingParameters",
    "GeneralError",
    "GlobalContainerContent",
    "GlobalContainerContent",
    "Icon",
    "ImplementationData",
    "ImplementationPlatform",
    "InheritEnvironment",
    "InvocationType",
    "IsLocalUser",
    "IsMappingRoutineCall",
    "IsSecurityRoutineCall",
    "KeepLoaded",
    "KeepName",
    "LastModificationTime",
    "LastStateChangeTime",
    "LowerLevel",
    "MappingType",
    "MembersOfRoles",
    "MessageText",
    "ObjectDescription",
    "Organization",
    "OrganizationType",
    "Origin",
    "Parameter",
    "Parameters",
    "PathAndFileName",
    "People",
    "PersonToNotify",
    "Priority",
    "ProcInstDescription",
    "ProcInstID",
    "ProcInstInputData",
    "ProcInstName",
    "ProcInstName",
    "ProcInstOutputData",
    "ProcInstParentID",
    "ProcInstParentName",
    "ProcInstState",
    "ProcInstSuspensionExpirationTime",
    "ProcInstTopLevelID",
    "ProcInstTopLevelName",
    "ProcTemplID",
    "ProcTemplName",
    "ProcTemplValidFromDate",
    "ProcessAdministrator",
    "ProcessInstance",
    "ProcessInstanceDelete",
    "ProcessInstanceDeleteResponse",
    "ProcessInstanceRestart",
    "ProcessInstanceRestartResponse",
    "ProcessInstanceResume",
    "ProcessInstanceResumeResponse",
    "ProcessInstanceStart",
    "ProcessInstanceStartResponse",
    "ProcessInstanceSuspend",
    "ProcessInstanceSuspendResponse",
    "ProcessInstanceTerminate",
    "ProcessInstanceTerminateResponse",
    "ProcessTemplateCreateAndStartInstance",
    "ProcessTemplateCreateAndStartInstanceResponse",
    "ProcessTemplateCreateInstance",
    "ProcessTemplateCreateInstanceResponse",
    "ProcessTemplateExecute",
    "ProcessTemplateExecuteResponse",
    "ProgramID",
    "ProgramInputData",
    "ProgramName",
    "ProgramOutputData",
    "ProgramOutputDataDefaults",
    "ProgramParameters",
    "ProgramRC",
    "Rc",
    "ResponseRequired",
    "Role",
    "RunInXTerm",
    "SecondActivityName",
    "SecondUserID",
    "ServiceName",
    "ServiceType",
    "StartInForeGround",
    "Starter",
    "SuspensionExpirationTime",
    "TerminateProgram",
    "TimeoutInterval",
    "TimeoutPeriod",
    "Timestamp",
    "UpperLevel",
    "UserContext",
    "UserID",
    "WfMessage",
    "WfMessageHeader",
    "WindowStyle",
    "WorkingDirectoryName",
    "_ACTIVITY",
    "_ACTIVITY_INFO",
    "_PROCESS",
    "_PROCESS_INFO",
    "_PROCESS_MODEL",
    "_RC",
    "null"
  };

  private static Hashtable names = new Hashtable(10 * elements.length);
}
